@@ -352,11 +352,10 @@ GRAPHENE_TEST_UNIT_BEGIN (matrix_2d_transforms)
 {
   graphene_matrix_t m1, m2;
   double xx, xy, yx, yy, x_0, y_0;
-  graphene_point3d_t tmp;
 
   graphene_matrix_init_identity (&m1);
   graphene_matrix_scale (&m1, 2.0, 2.0, 1.0);
-  graphene_matrix_translate (&m1, graphene_point3d_init (&tmp, 0.5, 0.5, 0.0));
+  graphene_matrix_translate (&m1, &GRAPHENE_POINT3D_INIT (0.5, 0.5, 0.0));
   if (g_test_verbose ())
     {
       g_test_message ("m1 -> translate(0.5, 0.5, 0.0) -> scale(2.0, 2.0)");
@@ -386,9 +385,9 @@ GRAPHENE_TEST_UNIT_BEGIN (matrix_2d_transforms)
   graphene_assert_fuzzy_equals (y_0, 0.5, 0.0001);
 
   graphene_matrix_init_identity (&m1);
-  graphene_matrix_translate (&m1, graphene_point3d_init (&tmp, 50, 50, 0));
+  graphene_matrix_translate (&m1, &GRAPHENE_POINT3D_INIT (50, 50, 0));
   graphene_matrix_rotate_z (&m1, 45.0);
-  graphene_matrix_translate (&m1, graphene_point3d_init (&tmp, -50, -50, 0));
+  graphene_matrix_translate (&m1, &GRAPHENE_POINT3D_INIT (-50, -50, 0));
   if (g_test_verbose ())
     {
       g_test_message ("m1 -> translate(50,50) -> rotate(45deg) -> translate(-50, -50)");
@@ -441,7 +440,7 @@ GRAPHENE_TEST_UNIT_BEGIN (matrix_interpolate)
   graphene_matrix_init_identity (&m1);
   graphene_matrix_init_identity (&m2);
 
-  graphene_matrix_translate (&m2, &(graphene_point3d_t) { 100.f, 100.f, 100.f });
+  graphene_matrix_translate (&m2, &GRAPHENE_POINT3D_INIT (100.f, 100.f, 100.f));
   g_assert_true (!graphene_matrix_is_2d (&m2));
 
   graphene_matrix_interpolate (&m1, &m2, 0.0, &mr);
@@ -450,12 +449,84 @@ GRAPHENE_TEST_UNIT_BEGIN (matrix_interpolate)
   graphene_matrix_interpolate (&m1, &m2, 1.0, &mr);
   graphene_assert_fuzzy_matrix_equal (&mr, &m2, 0.1);
 
-  graphene_matrix_init_translate (&m3, &(graphene_point3d_t) { 50.f, 50.f, 50.f });
+  graphene_matrix_init_translate (&m3, &GRAPHENE_POINT3D_INIT (50.f, 50.f, 50.f));
   graphene_matrix_interpolate (&m1, &m2, 0.5, &mr);
   graphene_assert_fuzzy_matrix_equal (&mr, &m3, 0.1);
 }
 GRAPHENE_TEST_UNIT_END
 
+GRAPHENE_TEST_UNIT_BEGIN (matrix_interpolate_perspective)
+{
+  graphene_matrix_t m1, m2, m3, mr;
+
+  graphene_matrix_init_identity (&m1);
+  graphene_matrix_perspective (&m1, 200, &m1);
+  graphene_matrix_init_identity (&m2);
+  graphene_matrix_perspective (&m2, 800, &m2);
+
+  graphene_matrix_interpolate (&m1, &m2, 0.0, &mr);
+  graphene_assert_fuzzy_matrix_equal (&mr, &m1, 0.1);
+
+  graphene_matrix_interpolate (&m1, &m2, 1.0, &mr);
+  graphene_assert_fuzzy_matrix_equal (&mr, &m2, 0.1);
+
+  graphene_matrix_init_identity (&m3);
+  graphene_matrix_perspective (&m3, 400, &m3);
+  graphene_matrix_interpolate (&m1, &m2, 0.5, &mr);
+  graphene_assert_fuzzy_matrix_equal (&mr, &m3, 0.1);
+}
+GRAPHENE_TEST_UNIT_END
+
+GRAPHENE_TEST_UNIT_BEGIN (matrix_multiply_self)
+{
+  graphene_matrix_t a, b, res, test;
+  float floats[16] = { 0, 0, 0, 2,
+                       0, 0, 2, 0,
+                       0, 2, 0, 0,
+                       2, 0, 0, 0 };
+
+  graphene_matrix_init_from_float (&a, floats);
+  graphene_matrix_init_from_float (&b, floats);
+  graphene_matrix_multiply (&a, &b, &res);
+
+  graphene_matrix_init_from_float (&test, floats);
+  graphene_matrix_multiply (&test, &b, &test);
+  graphene_assert_fuzzy_matrix_equal (&test, &res, G_MINDOUBLE);
+
+  graphene_matrix_init_from_float (&test, floats);
+  graphene_matrix_multiply (&a, &test, &test);
+  graphene_assert_fuzzy_matrix_equal (&test, &res, G_MINDOUBLE);
+
+  graphene_matrix_init_from_float (&test, floats);
+  graphene_matrix_multiply (&test, &test, &test);
+  graphene_assert_fuzzy_matrix_equal (&test, &res, G_MINDOUBLE);
+}
+GRAPHENE_TEST_UNIT_END
+
+GRAPHENE_TEST_UNIT_BEGIN (matrix_to_2d)
+{
+  graphene_matrix_t matrix;
+  float f[16];
+  guint i;
+  bool valid[16] = {
+    true,  true,  false, false,
+    true,  true,  false, false,
+    false, false, false, false,
+    true,  true,  false, false };
+
+  for (i = 0; i < 16; i++)
+    {
+      /* Take the identity matrix and change one member to a different value */
+      graphene_matrix_init_identity (&matrix);
+      graphene_matrix_to_float (&matrix, f);
+      f[i] = 0.5f;
+      graphene_matrix_init_from_float (&matrix, f);
+
+      g_assert_cmpint (graphene_matrix_to_2d (&matrix, NULL, NULL, NULL, NULL, NULL, NULL), ==, valid[i]);
+    }
+}
+GRAPHENE_TEST_UNIT_END
+
 GRAPHENE_TEST_UNIT_BEGIN (matrix_2d_interpolate)
 {
   graphene_matrix_t m1, m2, m3, mr;
@@ -463,20 +534,65 @@ GRAPHENE_TEST_UNIT_BEGIN (matrix_2d_interpolate)
   graphene_matrix_init_identity (&m1);
   graphene_matrix_init_identity (&m2);
 
-  graphene_matrix_translate (&m2, &(graphene_point3d_t) { 100.f, 100.f, 0.f });
+  graphene_matrix_translate (&m2, &GRAPHENE_POINT3D_INIT (100.f, 100.f, 0.f));
 
   g_assert_true (graphene_matrix_is_2d (&m1));
   g_assert_true (graphene_matrix_is_2d (&m2));
 
   graphene_matrix_interpolate (&m1, &m2, 0.0, &mr);
-  graphene_assert_fuzzy_matrix_equal (&mr, &m1, 0.01);
+  graphene_assert_fuzzy_matrix_equal (&mr, &m1, 0.01f);
 
   graphene_matrix_interpolate (&m1, &m2, 1.0, &mr);
-  graphene_assert_fuzzy_matrix_equal (&mr, &m2, 0.01);
+  graphene_assert_fuzzy_matrix_equal (&mr, &m2, 0.01f);
 
-  graphene_matrix_init_translate (&m3, &(graphene_point3d_t) { 50.f, 50.f, 0.f });
+  graphene_matrix_init_translate (&m3, &GRAPHENE_POINT3D_INIT (50.f, 50.f, 0.f));
   graphene_matrix_interpolate (&m1, &m2, 0.5, &mr);
-  graphene_assert_fuzzy_matrix_equal (&mr, &m3, 0.01);
+  graphene_assert_fuzzy_matrix_equal (&mr, &m3, 0.01f);
+}
+GRAPHENE_TEST_UNIT_END
+
+GRAPHENE_TEST_UNIT_BEGIN (matrix_2d_transform_bound)
+{
+  graphene_matrix_t m1, m2;
+  graphene_rect_t r, r2, res;
+
+  graphene_matrix_init_identity (&m1);
+  graphene_matrix_init_translate (&m2, &GRAPHENE_POINT3D_INIT (100.f, 100.f, 0.f));
+
+  graphene_rect_init (&r, 0.f, 0.f, 50.f, 50.f);
+  graphene_rect_init (&r2, 100.f, 100.f, 50.f, 50.f);
+
+  graphene_matrix_transform_bounds (&m1, &r, &res);
+  graphene_assert_fuzzy_rect_equal (&res, &r, 0.01f);
+
+  graphene_matrix_transform_bounds (&m2, &r, &res);
+  graphene_assert_fuzzy_rect_equal (&res, &r2, 0.01f);
+}
+GRAPHENE_TEST_UNIT_END
+
+GRAPHENE_TEST_UNIT_BEGIN (matrix_3d_transform_point)
+{
+  graphene_matrix_t m;
+  graphene_point_t p;
+  graphene_point3d_t p3;
+  graphene_vec3_t v;
+
+  graphene_matrix_init_translate (&m, &GRAPHENE_POINT3D_INIT (50.f, 70.f, 0.f));
+
+  g_test_message ("mat(translation) * point(zero) = point(translation)");
+  graphene_point_init (&p, 0.f, 0.f);
+  graphene_matrix_transform_point (&m, &p, &p);
+  graphene_assert_fuzzy_point_equal (&p, &GRAPHENE_POINT_INIT (50.f, 70.f), 0.01f);
+
+  g_test_message ("mat(translation) * point3d(zero) = point3d(translation)");
+  graphene_point3d_init (&p3, 0.f, 0.f, 0.f);
+  graphene_matrix_transform_point3d (&m, &p3, &p3);
+  graphene_assert_fuzzy_point3d_equal (&p3, &GRAPHENE_POINT3D_INIT (50.f, 70.f, 0.0f), 0.01f);
+
+  g_test_message ("mat(translation) * vec3(zero) = vec3(zero)");
+  graphene_vec3_init (&v, 0.f, 0.f, 0.f);
+  graphene_matrix_transform_vec3 (&m, &v, &v);
+  graphene_assert_fuzzy_vec3_equal (&v, graphene_vec3_zero (), 0.01f);
 }
 GRAPHENE_TEST_UNIT_END
 
@@ -490,8 +606,13 @@ GRAPHENE_TEST_SUITE (
   GRAPHENE_TEST_UNIT ("/matrix/look_at", matrix_look_at)
   GRAPHENE_TEST_UNIT ("/matrix/invert", matrix_invert)
   GRAPHENE_TEST_UNIT ("/matrix/interpolate", matrix_interpolate)
+  GRAPHENE_TEST_UNIT ("/matrix/interpolate-perspective", matrix_interpolate_perspective)
+  GRAPHENE_TEST_UNIT ("/matrix/multiply_self", matrix_multiply_self)
+  GRAPHENE_TEST_UNIT ("/matrix/to-2d", matrix_to_2d)
   GRAPHENE_TEST_UNIT ("/matrix/2d/identity", matrix_2d_identity)
   GRAPHENE_TEST_UNIT ("/matrix/2d/transforms", matrix_2d_transforms)
   GRAPHENE_TEST_UNIT ("/matrix/2d/round-trip", matrix_2d_round_trip)
   GRAPHENE_TEST_UNIT ("/matrix/2d/interpolate", matrix_2d_interpolate)
+  GRAPHENE_TEST_UNIT ("/matrix/2d/transform_bound", matrix_2d_transform_bound)
+  GRAPHENE_TEST_UNIT ("/matrix/3d/transform_point", matrix_3d_transform_point)
 )
